----------------------------------------
-- copy a table and all its subtables --
----------------------------------------
local function tcopy(intable,outtable,overwrite,written)
	--written is an internal table to track references to already copied tables and keep those intact
	--create outtable if it doesn't exist
	outtable = outtable or {}
	--create written if this is the first call (user call)
	written = written or {}
	--add copied table to written
	written[intable] = outtable
	--loop through data
	for key, data in pairs(intable) do
		--if overwrite or not exists, write value
		if (outtable[key] == nil) or overwrite then
			if type(data) == "table" then
				outtable[key] = written[data] or tcopy(data,nil,nil,written) -- we don't pass overwrite because the table copied to is freshly created anyways
			else
				outtable[key] = data
			end
		end
	end
	return outtable
end

local function isnpc(guid)
	return (guid:sub(5,5) == "3")
end

local function npcid(guid)
	return tonumber("0x"..guid:sub(7,10))
end

local EnemySwing = CreateFrame("Frame","EnemySwing")

local defaults = {
	savedSwingTimers = {},
	point = {
		x=0,
		y=0,
		p="CENTER",
		r="CENTER",
	},
	maxSwing = 10,
	unlock = false,
	width = 200,
	height = 20,
}

local lastSwingTimes = {
}

local targetBar = CreateFrame("Frame",nil,UIParent)
targetBar:SetMovable(true)
targetBar.background = targetBar:CreateTexture(nil,"BACKGROUND")
targetBar.background:SetAllPoints()
targetBar.background:SetTexture(0.8,0.8,0.8)
targetBar.background:SetAlpha(0.3)
targetBar.bar = targetBar:CreateTexture(nil,"ARTWORK")
targetBar.bar:SetPoint("TOPLEFT")
targetBar.bar:SetPoint("BOTTOMLEFT")
targetBar.bar:SetTexture(0.5,0.5,0.5)
targetBar.text = targetBar:CreateFontString(nil,"OVERLAY")
targetBar.text:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
targetBar.text:SetPoint("RIGHT",-10,0)
targetBar.text:SetTextColor(1,1,1)
targetBar.text2 = targetBar:CreateFontString(nil,"OVERLAY")
targetBar.text2:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
targetBar.text2:SetPoint("LEFT",10,0)
targetBar.text2:SetTextColor(1,1,1)
targetBar.movable = CreateFrame("Frame",nil,UIParent)
targetBar.movable:SetAllPoints(targetBar)
targetBar.movable.UpdateTex = function(self) self.tex[targetBar:IsShown() and "Hide" or "Show"](self.tex) end
targetBar.movable.tex = targetBar.movable:CreateTexture(nil,"OVERLAY")
targetBar.movable.tex:SetAlpha(0.7)
targetBar.movable.tex:SetTexture(1,1,1)
targetBar.movable.tex:SetAllPoints()
targetBar.movable:EnableMouse(true)
targetBar.movable:SetScript("OnMouseDown",function() targetBar:StartMoving() end)
targetBar.movable:SetScript("OnMouseUp",function() targetBar:StopMovingOrSizing() EnemySwing.db.point.p, _, EnemySwing.db.point.r, EnemySwing.db.point.x, EnemySwing.db.point.y = targetBar:GetPoint() end)
targetBar:SetScript("OnShow",function(s) s.movable.tex:Hide() end)
targetBar:SetScript("OnHide",function(s) s.movable.tex:Show() end)
targetBar.Update = function(self)
	if self.value >= self.maxValue then
		self:Hide()
		return
	end
	if self.value > 0 then
		self.bar:Show()
		self.bar:SetWidth((1-(self.value/self.maxValue))*self:GetWidth())
	else
		self.bar:Hide()
	end
	self.text:SetText(math.floor((self.maxValue-self.value)*10)/10)
end
targetBar:Hide()
targetBar.set = function(self,value)
	if not value then
		self:Hide()
	else
		self.value = value
		if self.maxValue then
			self:Show()
			self:Update()
		end
	end
end
targetBar.max = function(self,newMax)
	if not newMax then
		self:Hide()
	else
		self.maxValue = newMax
		if self.value then
			self:Show()
			self:Update()
		end
	end
end
targetBar:SetScript("OnUpdate",function(self)
	if not self.lastSwing then
		return
	end
	self:set(GetTime()-self.lastSwing)
end)

function EnemySwing:Swing(guid)
	local npc = npcid(guid)
	if npc then
		--print("Swing by NPC ID #"..npc)
		local curTime = GetTime()
		local lastTime = lastSwingTimes[guid]
		lastSwingTimes[guid] = curTime
		if lastTime then
			local swingTimer = curTime-lastTime
			if swingTimer > self.db.maxSwing then
				--print("Swing timer for "..npc.." discarded - above "..self.db.maxSwing.." sec")
			else
				self.db.savedSwingTimers[npc] = swingTimer
				if (UnitGUID("target") == guid) then
					targetBar:max(swingTimer)
					targetBar.lastSwing = curTime
					targetBar:Show()
				end
			end
		end	
	end
end

EnemySwing:SetScript("OnEvent",function(self,event,...)
	if event == "ADDON_LOADED" then
		if (...) == "EnemySwing" then
			EnemySwing_DB = tcopy(defaults,EnemySwing_DB)
			self.db = EnemySwing_DB
			if not self.db.unlock then
				targetBar.movable:Hide()
			end
			targetBar:SetSize(self.db.width,self.db.height)
			targetBar:SetPoint(self.db.point.p, UIParent, self.db.point.r, self.db.point.x, self.db.point.y)
			targetBar.movable:UpdateTex()
		end
	end
	if event == "COMBAT_LOG_EVENT_UNFILTERED" then
		local stype = select(2,...)
		if ((stype == "SWING_DAMAGE") or (stype == "SWING_MISSED")) then
			local guid = select(4,...)
			if isnpc(guid) then
				EnemySwing:Swing(guid)
			end
		end
	end
	if event == "PLAYER_TARGET_CHANGED" then
		local tGuid = UnitGUID("target")
		if tGuid then
			targetBar.text2:SetText((UnitName("target")))
			targetBar.lastSwing = lastSwingTimes[tGuid]
			targetBar:max(self.db.savedSwingTimers[npcid(tGuid)])
		end
	end
end)
EnemySwing:RegisterEvent("ADDON_LOADED")
EnemySwing:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
EnemySwing:RegisterEvent("PLAYER_TARGET_CHANGED")

SlashCmdList["ENEMYSWING"]=function(input)
	local key, value = input:match("(%S+) ?(.*)")
	if key == "maxSwing" then
		EnemySwing.db.maxSwing = tonumber(value) or defaults.maxSwing
		print("EnemySwing will now only saved swing timers below |cff8888ff"..EnemySwing.db.maxSwing.."|r seconds")
	elseif key == "unlock" then
		EnemySwing.db.unlock = (not EnemySwing.db.unlock)
		if EnemySwing.db.unlock then
			targetBar.movable:Show()
			targetBar.movable:UpdateTex()
			print("EnemySwing bar is now |cff00ff00UNLOCKED|r")
		else
			targetBar.movable:Hide()
			print("EnemySwing bar is now |cffff0000LOCKED|r")
		end
	elseif key == "height" then
		EnemySwing.db.height = tonumber(value) or defaults.height
		targetBar:SetHeight(EnemySwing.db.height)
	elseif key == "width" then
		EnemySwing.db.width = tonumber(value) or defaults.width
		targetBar:SetWidth(EnemySwing.db.width)
	else
		print("EnemySwing command parameters:")
		print("/enemyswing |cff8888ffparameter|r OR /enemyswing |cff8888ffparameter|r value")
		print("[|cff8888ffunlock|r] Bar movable: "..tostring(EnemySwing.db.unlock))
		print("[|cff8888ffmaxSwing|r] Maximum swing timer: "..tostring(EnemySwing.db.maxSwing))
		print("[|cff8888ffwidth|r] Bar width: "..tostring(EnemySwing.db.width))
		print("[|cff8888ffheight|r] Bar height: "..tostring(EnemySwing.db.height))
	end
end
SLASH_ENEMYSWING1 = "/enemyswing"